@@ -1,5 +1,5 @@
 # plugin/plugin_base.py
-# Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -54,7 +54,7 @@ def setup_options(make_option):
     make_option("--db", action="append", type="string", dest="db",
                 help="Use prefab database uri. Multiple OK, "
                 "first one is run by default.")
-    make_option('--dbs', action='callback', callback=_list_dbs,
+    make_option('--dbs', action='callback', zeroarg_callback=_list_dbs,
                 help="List available prefab dbs")
     make_option("--dburi", action="append", type="string", dest="dburi",
                 help="Database uri.  Multiple OK, "
@@ -63,6 +63,11 @@ def setup_options(make_option):
                 help="Drop all tables in the target database first")
     make_option("--backend-only", action="store_true", dest="backend_only",
                 help="Run only tests marked with __backend__")
+    make_option("--nomemory", action="store_true", dest="nomemory",
+                help="Don't run memory profiling tests")
+    make_option("--postgresql-templatedb", type="string",
+                help="name of template database to use for Postgresql "
+                     "CREATE DATABASE (defaults to current database)")
     make_option("--low-connections", action="store_true",
                 dest="low_connections",
                 help="Use a low number of distinct connections - "
@@ -177,6 +182,7 @@ def post_begin():
     warnings.setup_filters()
 
 
+
 def _log(opt_str, value, parser):
     global logging
     if not logging:
@@ -227,6 +233,12 @@ def _setup_options(opt, file_config):
     options = opt
 
 
+@pre
+def _set_nomemory(opt, file_config):
+    if opt.nomemory:
+        exclude_tags.add("memory_intensive")
+
+
 @pre
 def _monkeypatch_cdecimal(options, file_config):
     if options.cdecimal:
@@ -266,7 +278,13 @@ def _engine_uri(options, file_config):
     if not db_urls:
         db_urls.append(file_config.get('db', 'default'))
 
+    config._current = None
     for db_url in db_urls:
+
+        if options.write_idents and provision.FOLLOWER_IDENT: # != 'master':
+            with open(options.write_idents, "a") as file_:
+                file_.write(provision.FOLLOWER_IDENT + " " + db_url + "\n")
+
         cfg = provision.setup_config(
             db_url, options, file_config, provision.FOLLOWER_IDENT)
 
@@ -407,12 +425,21 @@ def want_method(cls, fn):
 def generate_sub_tests(cls, module):
     if getattr(cls, '__backend__', False):
         for cfg in _possible_configs_for_cls(cls):
-            name = "%s_%s_%s" % (cls.__name__, cfg.db.name, cfg.db.driver)
+            orig_name = cls.__name__
+
+            # we can have special chars in these names except for the
+            # pytest junit plugin, which is tripped up by the brackets
+            # and periods, so sanitize
+
+            alpha_name = re.sub('[_\[\]\.]+', '_', cfg.name)
+            alpha_name = re.sub('_+$', '', alpha_name)
+            name = "%s_%s" % (cls.__name__, alpha_name)
             subcls = type(
                 name,
                 (cls, ),
                 {
-                    "__only_on__": ("%s+%s" % (cfg.db.name, cfg.db.driver)),
+                    "_sa_orig_cls_name": orig_name,
+                    "__only_on_config__": cfg
                 }
             )
             setattr(module, name, subcls)
@@ -441,6 +468,12 @@ def _restore_engine():
     config._current.reset(testing)
 
 
+def final_process_cleanup():
+    engines.testing_reaper._stop_test_ctx_aggressive()
+    assertions.global_cleanup_assertions()
+    _restore_engine()
+
+
 def _setup_engine(cls):
     if getattr(cls, '__engine_options__', None):
         eng = engines.testing_engine(options=cls.__engine_options__)
@@ -451,11 +484,8 @@ def before_test(test, test_module_name, test_class, test_name):
 
     # like a nose id, e.g.:
     # "test.aaa_profiling.test_compiler.CompileTest.test_update_whereclause"
-    name = test_class.__name__
 
-    suffix = "_%s_%s" % (config.db.name, config.db.driver)
-    if name.endswith(suffix):
-        name = name[0:-(len(suffix))]
+    name = getattr(test_class, '_sa_orig_cls_name', test_class.__name__)
 
     id_ = "%s.%s.%s" % (test_module_name, name, test_name)
 
@@ -481,6 +511,9 @@ def _possible_configs_for_cls(cls, reasons=None):
             if not spec(config_obj):
                 all_configs.remove(config_obj)
 
+    if getattr(cls, '__only_on_config__', None):
+        all_configs.intersection_update([cls.__only_on_config__])
+
     if hasattr(cls, '__requires__'):
         requirements = config.requirements
         for config_obj in list(all_configs):
@@ -521,24 +554,20 @@ def _do_skips(cls):
                 )
 
     if not all_configs:
-        if getattr(cls, '__backend__', False):
-            msg = "'%s' unsupported for implementation '%s'" % (
-                cls.__name__, cls.__only_on__)
-        else:
-            msg = "'%s' unsupported on any DB implementation %s%s" % (
-                cls.__name__,
-                ", ".join(
-                    "'%s(%s)+%s'" % (
-                        config_obj.db.name,
-                        ".".join(
-                            str(dig) for dig in
-                            config_obj.db.dialect.server_version_info),
-                        config_obj.db.driver
-                    )
-                  for config_obj in config.Config.all_configs()
-                ),
-                ", ".join(reasons)
-            )
+        msg = "'%s' unsupported on any DB implementation %s%s" % (
+            cls.__name__,
+            ", ".join(
+                "'%s(%s)+%s'" % (
+                    config_obj.db.name,
+                    ".".join(
+                        str(dig) for dig in
+                        exclusions._server_version(config_obj.db)),
+                    config_obj.db.driver
+                )
+              for config_obj in config.Config.all_configs()
+            ),
+            ", ".join(reasons)
+        )
         config.skip_test(msg)
     elif hasattr(cls, '__prefer_backends__'):
         non_preferred = set()
